# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, you can obtain one at http://mozilla.org/MPL/2.0/.
import logging

import pytest
from pyspark.sql import SparkSession


@pytest.fixture(scope="session")
def spark():
    spark = (
        SparkSession
        .builder
        .master("local")
        .appName("mozanalysis_test")
        .getOrCreate()
    )

    logger = logging.getLogger("py4j")
    logger.setLevel(logging.ERROR)

    yield spark
    spark.stop()


@pytest.fixture(scope="session")
def spark_context(spark):
    return spark.sparkContext



import numpy as np
from functools import partial


def _resample(iteration, stat_fn, broadcast_data):
    np.random.seed(iteration)
    n = len(broadcast_data.value)
    randints = np.random.randint(0, n, n)
    return stat_fn(broadcast_data.value[randints])


def _percentile(data, confidence_level):
    """Returns low and high percentile based on provided confidence level"""
    confidence_margin = 0.5 * (1.0 - confidence_level)
    confidence_low = 0.0 + confidence_margin
    confidence_high = 1.0 - confidence_margin

    return {
        'confidence_low': np.percentile(data, confidence_low * 100),
        'confidence_high': np.percentile(data, confidence_high * 100),
    }


def bootstrap(sc, data, stat_fn, num_iterations=2000, confidence_level=0.95):
    """Returns a bootstrapped confidence interval

    Given an array of data, returns the bootstrapped confidence interval based
    on the confidence level provided. This uses pyspark and distributes the data
    to the nodes for efficiency.

    NOTE: All the data, plus one replicate, should fit into memory.

    Arguments:
    sc : The spark context.
    data : The data as a list or numpy array.
    stat_fn : A callable that computes the statistic.
    num_iterations : The number of bootstrap iterations to perform.
    confidence_level : The confidence level desired (between 0 and 1)

    Returns a dictionary with the keys::

        ["calcualted_value", "confidence_low", "confidence_high"]

    """
    if not type(data) == np.ndarray:
        data = np.array(data)

    # Broadcast the data as read-only to the clusters.
    broadcast_data = sc.broadcast(data)

    f = partial(_resample, stat_fn=stat_fn, broadcast_data=broadcast_data)
    stats = sc.parallelize(range(num_iterations)).map(f).collect()

    broadcast_data.unpersist()

    p = _percentile(stats, confidence_level)

    return {
        'calculated_value': stat_fn(data),
        'confidence_low': p['confidence_low'],
        'confidence_high': p['confidence_high'],
    }



import mock
import numpy as np

from mozanalysis.bootstrap import bootstrap, _percentile


# Generate a normalized distribution around a known mean.
MEAN = 15.0
TOLERANCE = 0.2
STDDEV = 0.1
DATA = np.random.normal(MEAN, STDDEV, 1000)


def test_percentile():
    data = np.arange(10)
    p = _percentile(data, 0.8)
    assert round(p['confidence_low'], 1) == 0.9
    assert round(p['confidence_high'], 1) == 8.1


def test_mean():
    assert MEAN - TOLERANCE < np.mean(DATA) < MEAN + TOLERANCE


def test_bootstrap_mean(spark_context):
    d = bootstrap(spark_context, DATA, np.mean, num_iterations=50,
                  confidence_level=0.95)
    # Check that the calculated mean from the samples are in expected range.
    assert MEAN - TOLERANCE < d['calculated_value'] < MEAN + TOLERANCE
    # Check that our confidence values are within expected ranges.
    low = np.mean(DATA) - 1.96 * STDDEV
    high = np.mean(DATA) + 1.96 * STDDEV
    assert low - TOLERANCE < d['confidence_low'] < low + TOLERANCE
    assert high - TOLERANCE < d['confidence_high'] < high + TOLERANCE


@mock.patch('mozanalysis.bootstrap._percentile')
def test_percentile_data(mocked, spark_context):
    # Test that we pass the bootstrap data to `_percentile`.
    data = np.arange(20)
    bootstrap(spark_context, data, np.mean, num_iterations=5)
    assert list(mocked.call_args[0][0]) != list(data)
